# -*- coding: utf8 -*-
'''
*******************************
***  DB BASE MODULE
*******************************
$RCSfile: base_dbmod.py,v $
$Revision: 1683 $
$Author: j.holetschek $
$Date: 2023-08-18 13:59:25 +0200 (Fr, 18. Aug 2023) $

The biocase.wrapper database base module to access datasources.
This is a base class for other dbmods that at least need to implement the connection method
if they are compliant to the Python DB API 2.0.'''

from biocase.wrapper.errorclasses import DbModuleNotSpecifiedError, SQLunicodeGenerationError, SQLError
from biocase.wrapper.sql.select import sqlSelect
from biocase.wrapper.sql.operators import getWhereObjTables

import logging, codecs
log = logging.getLogger("pywrapper.dbmod")


# Introduced when porting to Python3: set own handler for encoding errors
# This is useful if the data contains unprintable control characters that used to break BioCASe (like vertical tab, for example)
def encoding_error_handler(e):
    log.warn("UnicodeDecodeError: '%s' codec can't decode byte %s from position %s to %s (%s); replacing with '?'",
             e.encoding, e.object[e.start:e.end], e.start, e.end, e.reason)
    return ("?", e.end)


codecs.register_error("strict", encoding_error_handler)
log.info("Error handler for encoding/decoding exceptions has been set")


class DBBasemodule(sqlSelect):
    __dbmod__ = 'DB base module'

    def __repr__(self):
        return ("%s v%s" % (self.__class__.__dbmod__, self.__class__.__version__))

    def __init__(self, psfObj):
        sqlSelect.__init__(self, psfObj)
        self.conn = self.createDBconnection()

    def __del__(self):
        # release db connection
        try:
            self.conn.close()
            log.info("successfully closed %s db connection." % self.__dbmod__)
        except:
            log.error("error when trying to close the %s db connection." % self.__dbmod__)

    def createDBconnection(self):
        raise DbModuleNotSpecifiedError

    def execSelect(self, selectList, whereObj=None, orderByList=[], limit=None, offset=None):
        '''execute a select statement.
        the selectList gives the wanted attribute/alias tuples,
        the optional whereObj the WHERE filter,
        the optional orderbyList the ORDER BY clause as attribute/alias tuples,
        the optional limit the LIMIT/TOP limitation as an integer.'''
        # get SQL string
        (sql, paras) = self.select(selectList, whereObj, limit=limit, orderByList=orderByList, offset=offset)
        # execute statement
        return self.execSQL(sql, paras)

    def fixAllIds(self, AllIds):
        return AllIds

    def execDistinct(self, selectList, whereObj=None):
        '''Select distinct statement. The default routine does a regular select and filters only unique records via Python.'''
        log.debug("Select Distinct List: " + str(selectList))
        # build SELECT clause
        select = "DISTINCT %s" % (', '.join([self.joinDBAttrObj(dbt) for dbt in selectList]))
        (sql, paras) = self.select(selectList, whereObj, selectReplacementString=select, orderByList=selectList)
        return [r[0] for r in self.execSQL(sql, paras)]

    def execCount(self, DBAttrObj, whereObj=None):
        '''SELECT COUNT(x) statement. The default routine does a regular select and counts the records via Python.
        If the filter refers to the root table or is none, we can use a simple select count (because there will be only distinct primary keys);
        otherwise, we have to manually count distinct PKs (because count distinct is not supported by all DBMS)'''
        recCount = -1
        whereTables = getWhereObjTables(whereObj)
        if len(whereTables) == 0 or (len(whereTables) == 1 and whereTables[0] == DBAttrObj.table):
            log.info('Only root table queried; using SELECT COUNT to get number of records.')
            select = self.apply_template('count', [self.joinDBAttrObj(DBAttrObj)])
            (sql, paras) = self.select([DBAttrObj], whereObj, selectReplacementString=select)
            res = self.execSQL(sql, paras)
            if len(res) == 0:
                recCount = 0
            else:
                recCount = res[0][0]
        else:
            if 'count-distinct' in self.templates:
                log.info('Other tables that root table used in query; using count distinct to get number of records.')
                select = self.apply_template('count-distinct', [self.joinDBAttrObj(DBAttrObj)])
                (sql, paras) = self.select([DBAttrObj], whereObj, selectReplacementString=select)
                res = self.execSQL(sql, paras)
                if len(res) == 0:
                    recCount = 0
                else:
                    recCount = res[0][0]
            else:
                log.info('Other tables that root table used in query, count distinct not supported by DBMS; getting list of root table PKs and counting distinct values manually.')
                (sql, paras) = self.select([DBAttrObj], whereObj)
                PKs = set([rec[0] for rec in self.execSQL(sql, paras)])
                recCount = len(PKs)

        # return result
        return recCount

    def execSQL(self, sql, paras=None):
        # create cursor
        self.c = self.conn.cursor()
        try:
            if paras is not None:
                try:
                    log.info("Executing SQL: '%s' with parameters: %s" % (sql, "'" + "', '".join([str(p) for p in paras]) + "'"))
                except:
                    log.info("Executing SQL: '%s'" % (sql))
                    log.debug("Could not log used SQL parameters")
            else:
                log.info("Executing SQL: '%s'" % (sql))
        except:
            self.c.close()
            raise SQLunicodeGenerationError()
        # EXECUTE SQL
        log.debug("start DB query")
        try:
            if paras is None:
                self.c.execute(sql)
            else:
                if self.trans.integrateParameters == 0 and self.trans.parameterPlaceholder != '?':
                    sql = sql.replace('?', '%s')
                self.c.execute(sql, paras)
        except:
            raise SQLError(sql, paras)

        tmp = self.c.fetchall()
        log.debug("end DB query")
        log.info("Hits: " + str(len(tmp)))
        self.c.close()
        return tmp

    def getTables(self):
        log.error("in base")
        return None

    def getColumns(self, table):
        log.error("in base")
        return None

    def verifyTables(self, tableList):
        tables = self.getTables()
        if tables is None:
            return []
        errTables = []
        dbTables = [t.lower() for t in tables]
        for confTable in [ct.lower() for ct in tableList]:
            if confTable not in dbTables and confTable not in errTables:
                errTables.append(confTable)
        return errTables

    def hasMetaInfo(self):
        return False
