# -*- coding: utf8 -*-
'''
DB MODULE - ODBC for TSQL
*******************************
***  DB MODULE - ODBC for TSQL
*******************************
$RCSfile: dbmod_odbc_tsql.py,v $
$Revision: 1683 $
$Author: j.holetschek $
$Date: 2023-08-18 13:59:25 +0200 (Fr, 18. Aug 2023) $

The biocase.wrapper database module to access MS SQL Server databases.
You will need the pyODBC module to connect to a MS SQL Server DB via ODBC.
'''

from biocase.wrapper.sql.base import sqlBase
from biocase.wrapper.dbmod.base_dbmod import DBBasemodule
from biocase.wrapper.errorclasses import DbModuleMissingError, ConnectionError

import logging
log = logging.getLogger("pywrapper.dbmod")

# uses inherited ANSI SQL apart from the following.
sqlBase.templates['attribute'] = '[#1].[#2]'
sqlBase.templates['table'] = '[#1]'
sqlBase.templates['select-limit'] = 'SELECT DISTINCT TOP #3 #1 FROM #2'
sqlBase.templates['count-distinct'] = 'COUNT(DISTINCT #1)'


class DBmodule(DBBasemodule):
    __dbmod__ = 'MS SQL Server ODBC module using pyODBC'

    def createDBconnection(self):
        try:
            import pyodbc as pyodbc
        except:
            raise DbModuleMissingError('pyODBC')

        self.__class__.__version__ = pyodbc.version
        loginParas = "DSN=%s;UID=%s;PWD=%s" % (self.psfObj.database, self.psfObj.user, self.psfObj.password)
        try:
            con = pyodbc.connect(loginParas)

            # Try to get SQL Server version and turn on ROWS/OFFSET syntax
            try:
                ver = con.cursor().execute("SELECT cast(SERVERPROPERTY('ProductVersion') as varchar)").fetchone()[0]
                self.__class__.__version__ += " (SQL Server version %s)" % ver
                if int(ver[:2]) >= 11:
                    log.info("SQL Server version 2012+ found (%s); using ROWS/OFFSET syntax for selects." % ver)
                    sqlBase.templates['select-limit-offset'] = 'SELECT DISTINCT #1 FROM #2 OFFSET #4 ROWS FETCH NEXT #3 ROWS ONLY'
            except:
                log.warn("Retrieval of SQL Server version failed; using standard syntax.")

        except Exception as ex:
            log.exception(ex)
            raise ConnectionError()

        # Last, return connection
        return con

    def getTables(self):
        tables = []
        try:
            self.c = self.conn.cursor()
            self.c.execute("SELECT name FROM sysobjects where type in ('U', 'V') and status >= 0 order by name")
            for t in self.c.fetchall():
                tables.append(t[0])
        except:
            log.exception("Couldn't retrieve table info")
        finally:
            self.c.close()

        return tables

    def getColumns(self, _table):
        columns = []
        try:
            self.c = self.conn.cursor()
            self.c.execute("SELECT COLUMN_NAME FROM INFORMATION_SCHEMA.COLUMNS where table_name = '%s'" % (_table))
            for c in self.c.fetchall():
                columns.append(c[0])
        except:
            log.exception("Couldn't retrieve column info for table " + _table)
        finally:
            self.c.close()
        return columns

    def hasMetaInfo(self):
        return True
