# -*- coding: utf8 -*-
'''DB MODULE - PSYCO-POSTGRESQL
*******************************
***  DB MODULE - PSYCO-POSTGRESQL
*******************************
$RCSfile: dbmod_psycopg.py,v $
$Revision: 1683 $
$Author: j.holetschek $
$Date: 2023-08-18 13:59:25 +0200 (Fr, 18. Aug 2023) $

The biocase.wrapper database module to access PostgreSQL databases.
You will need the psycopg module to connect to a PostgreSQL DB.
'''

from biocase.wrapper.errorclasses import DbModuleMissingError, ConnectionError
from biocase.wrapper.sql.base import sqlBase
from biocase.wrapper.dbmod.base_dbmod import DBBasemodule

import logging
log = logging.getLogger("pywrapper.dbmod.postgres")

# uses inherited ANSI SQL apart from the following.
sqlBase.templates['attribute'] = '"#1"."#2"'
sqlBase.templates['table'] = '"#1"'
sqlBase.binaryCOP['like'] = "ILIKE"
sqlBase.templates['select-limit-offset'] = 'SELECT DISTINCT #1 FROM #2 LIMIT #3 offset #4'
sqlBase.templates['count-distinct'] = 'COUNT(DISTINCT #1)'


class DBmodule(DBBasemodule):
    __dbmod__ = 'Postgres DB module using psycopg2'

    def __init__(self, psfObj):
        DBBasemodule.__init__(self, psfObj)
        self.trans.parameterPlaceholder = '%s'

    def createDBconnection(self):
        try:
            import psycopg2
        except:
            raise DbModuleMissingError('psycopg2')

        self.__class__.__version__ = psycopg2.__version__
        try:
            # Extract port if given in IP field; otherwise, use default
            host, port = self.psfObj.dbIP, 5432
            if ":" in host:
                (host, _, port) = host.partition(":")
            # Connect
            loginParas = "host=%s port=%s user=%s dbname=%s " % (host, port, self.psfObj.user, self.psfObj.database)
            if len(self.psfObj.password) > 0:
                loginParas += "password=%s" % (self.psfObj.password)
            return psycopg2.connect(loginParas)
        except:
            raise ConnectionError()

    def getTables(self):
        if self.conn.server_version >= 80000:
            try:
                self.c = self.conn.cursor()
                self.c.execute("select table_name from information_schema.tables WHERE table_schema NOT IN ('pg_catalog', 'information_schema') AND table_type IN ('BASE TABLE', 'VIEW')")
                return [t[0] for t in self.c.fetchall()]
            except:
                log.exception("Couldn't retrieve table info")

    def getColumns(self, table):
        if self.conn.server_version >= 80000:
            try:
                self.c = self.conn.cursor()
                self.c.execute("SELECT column_name FROM information_schema.columns WHERE table_name = '%s'" % table)
                return [t[0] for t in self.c.fetchall()]
            except:
                log.exception("Couldn't retrieve column info for table " + table)

    def hasMetaInfo(self):
        try:
            return True if self.conn.server_version >= 80000 else False
        except:
            return False

    #===========================================================================
    # def verifyTables(self, tableList):
    #     # Overwrite this method to prevent error messages for configurations of Python2 installations.
    #     # The Python 2 BPS did not include the schema into table names; so a table PUBLIC.meta was stored as 'meta' (which also works; the PUBLIC schema is optional)
    #     # In order to prevent error messages, check if a table that wasn't found exists in the PUBLIC schema
    #     tables = self.getTables()
    #     if tables is None:
    #         return []
    #     errTables = []
    #     dbTables = [t.lower() for t in tables]
    #     for t in [ct.lower() for ct in tableList]:
    #         if t not in dbTables and ('public.' + t) not in dbTables and t not in errTables:
    #             errTables.append(t)
    #     return errTables
    #===========================================================================
